@@ -17,6 +17,10 @@
 #ifndef SIMPLE_PERF_PERF_EVENT_H_
 #define SIMPLE_PERF_PERF_EVENT_H_
 
+#if defined(USE_BIONIC_UAPI_HEADERS)
+#include <uapi/linux/perf_event.h>
+#else
 #include <linux/perf_event.h>
+#endif
 
 #endif  // SIMPLE_PERF_PERF_EVENT_H_
